@@ -8,10 +8,12 @@ import asyncio
 import hashlib
 import json
 import logging
+import re
 import socket
 
 from http import HTTPStatus
 from pathlib import Path
+from time import time
 from typing import Any, Dict, Generator, NamedTuple
 
 import aiofiles
@@ -44,7 +46,7 @@ pytestmark = [
     #         Compensate by specifying the namespace explicitly here, and it get_test_data().
     #       * Order of replication should be manifests, manifest lists, then tags.
     pytest.mark.push_image(
-        # "library/busybox:1.30.1",
+        "library/busybox:1.30.1",
         "library/busybox@sha256:4fe8827f51a5e11bb83afa8227cbccb402df840d32c6b633b7ad079bc8144100",
         "library/busybox@sha256:abc043b5132f825e44eefffc35535b1f24bd3f1bb60b11943863563a46795fdc",
         "library/busybox@sha256:07717dd5f074de0cf4f7ca8f635cb63aef63d789f15a22ab482a3d27a0a1f881",
@@ -54,7 +56,7 @@ pytestmark = [
         "library/busybox@sha256:a09f03056efb5d3facb5077a9e58e83e9bba74ad4d343b2afa92c70b5ae01e2b",
         "library/busybox@sha256:0b671b6a323d86aa6165883f698b557ca257c3a3ffa1e3152ffb6467e7ac11b3",
         "library/busybox@sha256:4b6ad3a68d34da29bf7c8ccb5d355ba8b4babcad1f99798204e7abb43e54ee3d",  # ManifestList
-        # "library/python:3.7.2-slim-stretch",
+        "library/python:3.7.2-slim-stretch",
         "library/python@sha256:0005ba40bf87e486d7061ca0112123270e4a6088b5071223c8d467db3dbba908",
         "library/python@sha256:09001905f918a977427cc6931a1cac84a8645b1ac2011fd3f40f625daf9a7fb1",
         "library/python@sha256:2d127b64fbb7a58ee8eb2c321f1bbd14548ab3191009cca7845b81155c9067bf",
@@ -220,7 +222,7 @@ async def docker_registry_client_async(
     ) as docker_registry_client_async:
         credentials = docker_registry_secure.auth_header["Authorization"].split()[1]
         await docker_registry_client_async.add_credentials(
-            docker_registry_secure.endpoint, credentials
+            credentials=credentials, endpoint=docker_registry_secure.endpoint
         )
 
         yield docker_registry_client_async
@@ -280,7 +282,7 @@ async def replicate_manifest_lists(docker_registry_secure: DockerRegistrySecure)
                 image_name.image
             )
             auth_header_src = await docker_registry_client_async._get_request_headers(
-                image_name, scope=scope
+                image_name=image_name, scope=scope
             )
             if not auth_header_src:
                 LOGGER.warning(
@@ -318,19 +320,37 @@ async def test___init__(docker_registry_client_async: DockerRegistryClientAsync)
 async def test_add_credentials(docker_registry_client_async: DockerRegistryClientAsync):
     """Test that credentials can be assigned."""
     endpoint = "endpoint"
-    credentials = "credentials"
-    await docker_registry_client_async.add_credentials(endpoint, credentials)
-    result = await docker_registry_client_async._get_credentials(endpoint)
+    credentials = f"credentials{time()}"
+    await docker_registry_client_async.add_credentials(
+        credentials=credentials, endpoint=endpoint
+    )
+    result = await docker_registry_client_async._get_credentials(endpoint=endpoint)
     assert result == credentials
 
 
+async def test_add_token(docker_registry_client_async: DockerRegistryClientAsync):
+    """Test that tokens can be assigned."""
+    endpoint = "endpoint"
+    for scope in [None, f"scope{time()}"]:
+        token = f"token{time()}"
+        await docker_registry_client_async.add_token(
+            endpoint=endpoint, scope=scope, token=token
+        )
+        result = await docker_registry_client_async._get_token(
+            endpoint=endpoint, scope=scope
+        )
+        assert result == token
+
+
 @pytest.mark.online
 async def test__get_auth_token_dockerhub():
     """Test that an authentication token can be retrieved for index.docker.io."""
     endpoint = Indices.DOCKERHUB
     # Note: Using default credentials store from the test environment
     async with DockerRegistryClientAsync() as docker_registry_client_async:
-        credentials = await docker_registry_client_async._get_credentials(endpoint)
+        credentials = await docker_registry_client_async._get_credentials(
+            endpoint=endpoint
+        )
         if credentials:
             token = await docker_registry_client_async._get_auth_token(
                 credentials=credentials,
@@ -362,7 +382,9 @@ async def test__get_auth_token_quay():
     endpoint = Indices.QUAY
     # Note: Using default credentials store from the test environment
     async with DockerRegistryClientAsync() as docker_registry_client_async:
-        credentials = await docker_registry_client_async._get_credentials(endpoint)
+        credentials = await docker_registry_client_async._get_credentials(
+            endpoint=endpoint
+        )
         if credentials:
             token = await docker_registry_client_async._get_auth_token(
                 credentials=credentials,
@@ -383,7 +405,9 @@ async def test__get_auth_token_redhat():
     endpoint = Indices.REDHAT
     # Note: Using default credentials store from the test environment
     async with DockerRegistryClientAsync() as docker_registry_client_async:
-        credentials = await docker_registry_client_async._get_credentials(endpoint)
+        credentials = await docker_registry_client_async._get_credentials(
+            endpoint=endpoint
+        )
         if credentials:
             token = await docker_registry_client_async._get_auth_token(
                 credentials=credentials,
@@ -416,10 +440,49 @@ async def test__get_credentials(
     docker_registry_client_async: DockerRegistryClientAsync, endpoint: str, auth: str
 ):
     """Test that credentials can be retrieved."""
-    result = await docker_registry_client_async._get_credentials(endpoint)
+    result = await docker_registry_client_async._get_credentials(endpoint=endpoint)
     assert result == auth
 
 
+@pytest.mark.parametrize(
+    "endpoint",
+    [
+        "com",
+        "com/",
+        "com:1234",
+        "com:1234/",
+        "google.com",
+        "google.com/",
+        "google.com:1234",
+        "google.com:1234/",
+        "www.google.com",
+        "www.google.com/",
+        "www.google.com:1234",
+        "www.google.com:1234/",
+        "https://com",
+        "https://com/",
+        "https://com:1234",
+        "https://com:1234/",
+        "https://google.com",
+        "https://google.com/",
+        "https://google.com:1234",
+        "https://google.com:1234/",
+        "https://www.google.com:1234",
+        "https://www.google.com:1234/",
+    ],
+)
+async def test__get_endpoint_pattern(endpoint: str):
+    """Tests that endpoints can be converted into regular expression patterns."""
+    pattern = await DockerRegistryClientAsync._get_endpoint_pattern(endpoint=endpoint)
+    assert pattern
+    pattern_raw = pattern.pattern
+    assert pattern_raw
+    assert pattern_raw.startswith("^")
+    assert pattern_raw.endswith("$")
+    _endpoint = endpoint.replace("https://", "").replace("/", "")
+    assert re.escape(_endpoint) in pattern_raw
+
+
 async def test__get_proxy(docker_registry_client_async: DockerRegistryClientAsync):
     """Test that a proxy can be retrieved."""
     endpoints = ["endpoint0.com", "endpoint1.com", "127.0.0.1"]
@@ -506,23 +569,19 @@ async def test__get_proxy(docker_registry_client_async: DockerRegistryClientAsyn
             assert result == url
 
 
-@pytest.mark.parametrize(
-    "endpoint,auth",
-    [
-        ("endpoint:port", "dXNlcm5hbWU6cGFzc3dvcmQ="),
-        ("endpoint2:port2", "dXNlcm5hbWUyOnBhc3N3b3JkMg=="),
-    ],
-)
 async def test__get_request_headers_basic_auth(
-    docker_registry_client_async: DockerRegistryClientAsync, endpoint: str, auth: str
+    docker_registry_client_async: DockerRegistryClientAsync,
+    docker_registry_secure: DockerRegistrySecure,
+    known_good_image: TypingKnownGoodImage,
 ):
     """Test request headers retrieval."""
-    image_name = ImageName("", endpoint=endpoint)
     existing_header = "existing-header"
     headers = await docker_registry_client_async._get_request_headers(
-        image_name, {existing_header: "1"}
+        headers={existing_header: "1"}, image_name=known_good_image.image_name
+    )
+    assert (
+        docker_registry_secure.auth_header["Authorization"] == headers["Authorization"]
     )
-    assert auth in headers["Authorization"]
     assert existing_header in headers
 
 
@@ -533,13 +592,93 @@ async def test__get_request_headers_token_anonymous():
     existing_header = "existing-header"
     async with DockerRegistryClientAsync() as docker_registry_client_async:
         headers = await docker_registry_client_async._get_request_headers(
-            image_name, {existing_header: "1"}
+            headers={existing_header: "1"}, image_name=image_name
         )
         assert headers["Authorization"].startswith("Bearer ")
         assert len(headers["Authorization"]) > 100
         assert existing_header in headers
 
 
+@pytest.mark.online
+async def test__get_token_dockerhub():
+    """Test that an authentication token can be retrieved for index.docker.io."""
+    endpoint = Indices.DOCKERHUB
+    # Note: Using default credentials store from the test environment
+    async with DockerRegistryClientAsync() as docker_registry_client_async:
+        credentials = await docker_registry_client_async._get_credentials(
+            endpoint=endpoint
+        )
+        if credentials:
+            token = await docker_registry_client_async._get_token(
+                credentials=credentials,
+                endpoint=endpoint,
+                scope=DockerAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                    "busybox"
+                ),
+            )
+            assert len(token) > 100
+        else:
+            pytest.skip(f"Unable to retrieve credentials for: {endpoint}")
+
+
+@pytest.mark.online
+async def test__get_token_dockerhub_anonymous():
+    """Test that an authentication token can be retrieved for index.docker.io anonymously."""
+    # Note: Using default credentials store from the test environment
+    async with DockerRegistryClientAsync() as docker_registry_client_async:
+        token = await docker_registry_client_async._get_token(
+            endpoint=Indices.DOCKERHUB,
+            scope=DockerAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format("busybox"),
+        )
+        assert len(token) > 100
+
+
+@pytest.mark.online
+async def test__get_token_quay():
+    """Test that an authentication token can be retrieved for quay.io."""
+    endpoint = Indices.QUAY
+    # Note: Using default credentials store from the test environment
+    async with DockerRegistryClientAsync() as docker_registry_client_async:
+        credentials = await docker_registry_client_async._get_credentials(
+            endpoint=endpoint
+        )
+        if credentials:
+            token = await docker_registry_client_async._get_token(
+                credentials=credentials,
+                endpoint=endpoint,
+                # scope=QuayAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                scope=DockerAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                    "crio/busybox"
+                ),
+            )
+            assert len(token) > 100
+        else:
+            pytest.skip(f"Unable to retrieve credentials for: {endpoint}")
+
+
+@pytest.mark.online
+async def test__get_token_redhat():
+    """Test that an authentication token can be retrieved for registry.redhat.io."""
+    endpoint = Indices.REDHAT
+    # Note: Using default credentials store from the test environment
+    async with DockerRegistryClientAsync() as docker_registry_client_async:
+        credentials = await docker_registry_client_async._get_credentials(
+            endpoint=endpoint
+        )
+        if credentials:
+            token = await docker_registry_client_async._get_token(
+                credentials=credentials,
+                endpoint=endpoint,
+                # scope=RedHatAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                scope=DockerAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                    "ocs4/ocs-rhel8-operator"
+                ),
+            )
+            assert len(token) > 100
+        else:
+            pytest.skip(f"Unable to retrieve credentials for: {endpoint}")
+
+
 @pytest.mark.parametrize(
     "endpoint,auth",
     [
@@ -553,9 +692,9 @@ async def test__load_credentials(
     """Test that credentials can be loaded from the credentials store."""
     await docker_registry_client_async._load_credentials()
     found = False
-    for key in docker_registry_client_async.credentials:
-        if endpoint in key:
-            assert docker_registry_client_async.credentials[key]["auth"] == auth
+    for pattern, credentials in docker_registry_client_async.credentials.items():
+        if pattern.fullmatch(endpoint):
+            assert credentials == auth
             found = True
     assert found
 
@@ -2123,5 +2262,85 @@ async def test_issue_25(
         assert response.result
 
 
+@pytest.mark.online
+async def test_issue_26():
+    """Test issue #26."""
+    endpoint = Indices.REDHAT
+    # Note: Using default credentials store from the test environment
+    for fallback_basic_auth in [True, False]:
+        for pattern in [r"^.*\.redhat\.io$", r"^.*\.io$", r"^.*$"]:
+            async with DockerRegistryClientAsync(
+                fallback_basic_auth=fallback_basic_auth
+            ) as docker_registry_client_async:
+                credentials = await docker_registry_client_async._get_credentials(
+                    endpoint=endpoint
+                )
+                if credentials:
+                    # Convert registry.redhat.io credentials into a pattern and remove all others ...
+                    docker_registry_client_async.credentials = {}
+                    await docker_registry_client_async.add_credentials(
+                        credentials=credentials, endpoint=re.compile(pattern)
+                    )
+                    assert len(docker_registry_client_async.credentials.keys()) == 1
+
+                    token = await docker_registry_client_async._get_token(
+                        credentials=credentials,
+                        endpoint=endpoint,
+                        # scope=RedHatAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                        scope=DockerAuthentication.SCOPE_REPOSITORY_PULL_PATTERN.format(
+                            "ocs4/ocs-rhel8-operator"
+                        ),
+                    )
+                    assert len(token) > 100
+                else:
+                    pytest.skip(f"Unable to retrieve credentials for: {endpoint}")
+
+
+@pytest.mark.online
+async def test_issue_27(
+    docker_registry_secure: DockerRegistrySecure, known_good_image: TypingKnownGoodImage
+):
+    """Test issue #27."""
+    credentials = docker_registry_secure.auth_header["Authorization"].split()[1]
+
+    image_names = list(get_identifier_map(known_good_image).keys())
+    image_name = image_names[0]
+
+    # connector_owner = True
+    docker_registry_client_async = DockerRegistryClientAsync(
+        ssl=docker_registry_secure.ssl_context
+    )
+    await docker_registry_client_async.add_credentials(
+        credentials=credentials, endpoint=docker_registry_secure.endpoint
+    )
+    LOGGER.debug("Checking manifest for: %s (#0) ...", image_name)
+    response = await docker_registry_client_async.head_manifest(image_name)
+    assert response.client_response
+    assert response.result
+    await docker_registry_client_async.close()
+
+    # connector_owner = False
+    docker_registry_client_async = DockerRegistryClientAsync(
+        client_session_kwargs={"connector_owner": False},
+        ssl=docker_registry_secure.ssl_context,
+    )
+    await docker_registry_client_async.add_credentials(
+        credentials=credentials, endpoint=docker_registry_secure.endpoint
+    )
+    LOGGER.debug("Checking manifest for: %s (#1) ...", image_name)
+    response = await docker_registry_client_async.head_manifest(image_name)
+    assert response.client_response
+    assert response.result
+    await docker_registry_client_async.close()
+    # Retry after close ...
+    LOGGER.debug("Checking manifest for: %s (#2) ...", image_name)
+    response = await docker_registry_client_async.head_manifest(image_name)
+    assert response.client_response
+    assert response.result
+
+    # Cleanup
+    docker_registry_client_async.client_session_kwargs["connector"].close()
+
+
 # TODO: Total image pull (with exists() checks)
 # TODO: Total image push
